/* pbmtodot.c - read a portable bitmap and produce dot-matrix printer graphics
**
** Copyright (C) 1992 by Ian Phillipps (ian@unipalm.co.uk)
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/

#include <stdio.h>
#ifdef __TURBOC__
#include <stdlib.h>
#endif
#include "pbm.h"

/*
This is like Termcap, but simplified. Each parameter name is one letter.
String parameter names are followed by a delimiter which
must not occur in the string.
Numerical parameters may be followed by "=" or ":".

<str>	::= <item> <whitespace> <str> | <item>
<item>	::= <parmname> <delimiter> <string> <delimiter> | <shortname> <number>
<parmname> ::=  c | d | e | f | l | m | s | u | x
<shortname> ::= b | i | p

These default to null or 1 unless stated.
l	Start line (usually contains %c)
e	End line (should not do a linefeed: defaults to \r)
f	Line feed (end of line: defaults to \n)
m	Line feed (between multiple passes)
s	Start Print (initialise printer)
x	Exit print (reset printer to normal, maybe do a pagefeed)
b	Number of bytes transmitted per pass
i	Interleave (number of passes with tiny line feed between each)
p	number of pins per pass (normally 8*by: defaults to 8)
d *	Non-zero if lsb is top row of dots
u *	Non-zero if first byte is at the bottom (fill Upwards)
	* = not implemented.
c	Comment (parameter is ignored)

Escape sequences

%c	Number of columns as two binary bytes, little-endian
	(This is the only variable that printers want that I know)
	If you need more of the same, hack "putcode". And tell me.
%%	Print a "%"

	The "\" escapes are provided for the benefit of PBMTODOT
	variable setters. If you want an escape in a compiled-in string,
	you must use \\e, to ensure that the runtime sees the "\".
	If you need to print a literal "\", you need four of them. Sorry.

\\	print a "\"
\r	Carriage Return
\n	Line feed
\e	Escape - NB This must be represented as \\e within a C string
\123	Octal number
	NB - the string must NOT contain a literal \0 when scanned, so if
	it is compiled-in to the program, a binary zero must be shown as
	\\0. Once readparm has been at the string, an embedded \0 is fine.

*/

/* defcon is the default control string used if no parameters or environment
   variables are specified */

char defcon[] = "l:\\e*\\047%c: f:\\eJ\030: b=3 p=24\
c:Gives 180 dpi resolution on NEC and Epson 24-print printers:";

/* Pairs of command-line parms and strings they produce. Because of the
   way pm_getparms operates, make sure that a string is not followed by
   one which is a leading substring of itself (e.g. -24x2 must come
   before -24) */

char *modes[][2] =
{
    {	"-24x2",
	/* 360 dpi for a typical 24-pin printer (NEC 2200): two passes
	   sets line spacing to 1/6" on exit */
	"s:\034\063\001: x:\\e2: l:\\e*\050%c: m:\n: \
	f:\n\\eJ\027: b=3 i=2 p=24\
	c:Gives 360 dpi resolution on NEC 24-print printers:"
    },
    {	"-24",
	defcon
    },
    {	"-48",
	/* 360 dpi resolution on a Canon bubblejet (known to work on BJ-10e) */
	/* Sorry, but we need to transmit a "\" character: hence \\\\ ! */
	"s:\\e[\\\\\4\\0\\0\\0h\1: l:\34CB\4\4%c: \
	f:\\eJ\60: p=48 b=6"
    },
    {	"-8x3",
	/* Epson 9-pin, three passes: gives 240 x 216 resolution */
	"l:\\eZ%c: f:\\eJ\25: m:\\eJ\1: i=3"
    },
    { 	"-8",
	/* Epson 9-pin (only 8 used) - bog-standard single density 60x72 resolution */
	"s:\\eA\010: l:\\eK%c:"
    },
    { (char*)0, (char*)0 } /* Keep this at the end! */
};

static char *st_start_line;
static char *st_end_line;
static char *linefeed;
static char *tinyfeed;
static char *st_start_print;
static char *st_end_print;
static short bytes;
static short passes; /* i */
static short pins;
static FILE *ofp;
static int cols;
static int cols_x_passes;

#ifdef __MSDOS__
/* Would you believe that MSDOS (version 5) won't send a ^Z to a printer
   via a filehandle?  No, nor did I until I wrote this program.
 */
#include <bios.h>
#include <conio.h>
#define PR_ERROR	0x29	/* Error bits in printer status */
#define PRINTER 	0	/* LPT1 */
/* send characters to LPT1 via the BIOS. I dare say that Dos call 5 would do
   the job as well, but you don't get a chance to print diagnostics */
int printc( int c )
{
	if ( ofp != stdout ) return putc( c, ofp );
	while( biosprint( 2, 0, PRINTER ) & PR_ERROR )
	{
		fprintf( stderr, "Printer error: correct then press any key\n");
		getch();
	}
	while( biosprint( 0, c, PRINTER ) & PR_ERROR )
	{
		fprintf( stderr, "Printer error: correct then press any key\n");
		getch();
	}
	return c;

}
#else /* you have a sensible operating system */
#define printc( c ) putc( (c), ofp )
#endif


/* Readparm reads a control string, and may destroy it in the process */
void readparm( string )
char *string;
{
    /* initialise variables */
    st_start_line = "::";
    st_end_line = ":\r:";
    linefeed = ":\n:";
    tinyfeed = "::";
    st_start_print = "::";
    st_end_print = "::";
    bytes = 1;
    passes = 1;
    pins = 8;
    /* read the string */
    while( *string )
    {
	char c = tolower( *string );
	short *ashort;
	char *dest, delim;

	string++;

	switch ( c )
	{
	case '\0':
	    return;

	case ' ': case '\t': case '\r': case '\n':
	    break;

	/* numerical parameters */
	case 'b': ashort = &bytes; goto decode_short;
	case 'i': ashort = &passes; goto decode_short;
	case 'p': ashort = &pins;
	decode_short:
	    while( *string == ':' || *string == '=' )
		string++;
	    *ashort = 0;
	    while( isdigit( *string ) )
		*ashort = *ashort * 10 + *string++ - '0';
	    break;

	case 'e': st_end_line = string;	    	goto decode_string;
	case 'f': linefeed = string;		goto decode_string;
	case 'l': st_start_line = string;	goto decode_string;
	case 'm': tinyfeed = string;		goto decode_string;
	case 's': st_start_print = string;	goto decode_string;
	case 'x': st_end_print = string;

	case 'c': /* no action needed for a comment */
	decode_string:
	    /* Decode the string in situ */
	    delim = *string++;
	    dest = string;
	    /* The code to overwrite the new copy of string relies
	       on the interpreted string being no longer than the
	       original */
	    for( ; ( *dest = c = *string++ ) != delim; dest++ )
	    {
		if (!c )
		    pm_error( "Delimiter '%c' missing from PBMTODOT string",
					delim );
		if ( c == '\\' ) switch ( *string )
		{
		case 'r':
		    *dest = '\r'; string++; break;

		case 'n':
		    *dest = '\n'; string++; break;

		case 'e':
		    *dest = '\033'; string++; break;

		default:
		    *dest = *string++; break;

		case '0': case '1': case '2': case '3':
		case '4': case '5': case '6': case '7':
		    *dest = *string++ - '0';
		    if ( '0' <= *string && *string <= '7' )
		    {
			*dest = (*dest << 3) + *string++ - '0';
			if ( '0' <= *string && *string <= '7' )
			    *dest = (*dest << 3) + *string++ - '0';
		    }
		}
	    }
	    break;

	case 'd': /* not implemented */
	case 'u': /* *not implemented */
	default:
	    pm_error( "unknown char %c in format", c );
	}
    }
}

/* Write a single control string to the printer or output */
void putcode( format )
    char *format;
{
    char delim;
	char c;
    for( delim = *format++; ( c = *format ) != delim; format++ )
    {
	if ( c == '%' )
	{
	    switch ( *++format )
	    {
	    case 'c':
		printc( cols & 255 );
		printc( cols >> 8 );
		break;
	    default:
		printc( *format );
	    }
	}
	else if ( printc( c ) == EOF )
		pm_error( "Output error" );
	}
}

/* Output array contains passes in pass-major order, so print out everything
   in the order it's stored */

void writepasses( startline )
	char *startline;
{
    int pass, col, byte;
    char *outbyte;
    for( pass = 0; ;  )
    {
	char *outcol;
	int maxcols = cols;		/* save value of columns */
	char *lastblank = outbyte = startline;	/* Last position to test for blank */
	cols = 0;
	/* Although this order of doing things looks slower, it probably isnt*/
	for( byte = 0; byte < bytes; byte++ )
	{
	    outcol = outbyte + maxcols;
	    while ( ( *--outcol == 0 ) && ( outcol >= lastblank ) )
		;
	    if ( outcol > lastblank )
	    {
		cols = lastblank - outbyte;
		lastblank = outcol;
	    }
	    outbyte += maxcols;
	    lastblank += maxcols;
	}

	outcol = startline;
	putcode ( st_start_line );
	for( col = 0; col < cols; col++ )
	    for( outbyte = outcol++, byte = 0; byte < bytes; byte++ )
		printc( *outbyte ), outbyte += maxcols;
	putcode ( st_end_line );
	cols = maxcols;
	if ( ++pass >= passes ) break;
	putcode( tinyfeed );
	startline += maxcols * bytes;
    }
    putcode( linefeed );

}

void main( argc, argv )
    int argc;
    char* argv[];
{
    FILE* ifp;
    bit *inrow, *inrowend;
    char* outarray;

    char *control;				/* Control string address */

    int rows, format;

    pbm_init( &argc, argv );

    argc--; argv++;

    control = getenv( "PBMTODOT" );
    if ( ! control ) control = defcon;

    if ( **argv == '-' )
    {
	int m = 0;
	for ( m = 0; *modes[m][0]; m++ )
	    if( pm_keymatch( *argv, modes[m][0], strlen( modes[m][0] ) ) )
	    {
		control = modes[m][1];
		argv++; argc--;
		break;
	    }
    }

    if ( argc > 2 || **argv == '-' )
	pm_usage( "[-8 -8x3 -24 -24x2 -48] [infile] [outfile]" );

    ifp = pm_openr( argc >= 1 ? argv[0] : "-" );
    readparm( control );
    ofp = argc >= 2 ? pm_openw( argv[1] ) : stdout;

    pbm_readpbminit(ifp, &cols, &rows, &format);
    cols_x_passes = cols * passes;

    /* allocate space for output */

    if( !( outarray = malloc( cols_x_passes * bytes ) ) )
	pm_error( "Out of memory" );

    inrow = pbm_allocrow(cols);
    inrowend = inrow + cols;

    /* Read in the file */
    putcode( st_start_print );
    /* here we have three loops: (inner) passes, pins and (outer) bytes */
    while( rows > 0 )
    {
	unsigned char mask = 0x80;
	char *startpass = outarray;
	int r;

	memset( outarray, 0, cols_x_passes * bytes );
	for( r = 0; r < pins; r++ )
	{
	    int p;
	    char *startcol = startpass;
	    for ( p = 0; p < passes; p++, startcol += cols * bytes )
	    {
		char *o = startcol;
		bit *in = inrow;
		pbm_readpbmrow(ifp, inrow, cols, format);
		for( ; in < inrowend; o++, in++ )
		    if( *in == PBM_BLACK )
			*o |= mask;
		/* breakout here ensures that the last few lines get printed */
		if ( --rows <= 0 )
		    goto endwriteloop; /* don't panic: just a double "break" */
	    }
	    if ( !( mask >>= 1 ) )
	    {
		mask = 0x80;
		startpass += cols;
	    }
	}
    endwriteloop:
	writepasses( outarray );
    }
    putcode( st_end_print );

    pm_close( ifp );

    exit( 0 );
    }
